# -*- coding: utf-8 -*-

'''Declarative metaclasses to incapsulate repeated code into base class. To
combine them you have to create a subclass. Example:

    TABLE_ARGS = {
        'mysql_engine': 'InnoDB',
        'mysql_default charset': 'utf8',
    }

    class Meta(AutoTableNameMeta, TableArgsMeta(TABLE_ARGS)):
        pass

    Base = declarative_base(name='Base', metaclass=Meta)

    class MyClass(Base):
        …

is equivalent to

    class MyClass(Base):
        __tablename__ = 'MyClass'
        __table_args__ = {
            'mysql_engine': 'InnoDB',
            'mysql_default charset': 'utf8',
        }
        …
'''

from sqlalchemy.ext import declarative
from ...utils.deprecation import deprecated


class AutoTableNameMeta(declarative.DeclarativeMeta):
    '''Declarative metaclass automatically giving the same name to table as
    mapped class. Example:

        Base = declarative_base(name='Base', metaclass=AutoTableNameMeta)

        class MyClass(Base):
            …

    is equivalent to

        Base = declarative_base(name='Base')

        class MyClass(Base):
            __tablename__ = 'MyClass'
            …
    '''

    def __init__(cls, name, bases, dict_):
        # Do not extend base class
        if '_decl_class_registry' not in cls.__dict__:
            # Lookup __dict__ and not attribute directly since __tablename__
            # should change on inheritance.
            if '__tablename__' not in cls.__dict__ and \
                    not cls.__dict__.get('__abstract__', False):
                cls.__tablename__ = name
            # XXX Commented untill we write the reason in comments
            #elif cls.__tablename__ is None:
            #    del cls.__tablename__
        super(AutoTableNameMeta, cls).__init__(name, bases, dict_)


def TableArgsMeta(table_args):
    '''Declarative metaclass automatically adding (merging) __table_args__ to
    mapped classes. Example:

        Meta = TableArgsMeta({
            'mysql_engine': 'InnoDB',
            'mysql_default charset': 'utf8',
        }

        Base = declarative_base(name='Base', metaclass=Meta)

        class MyClass(Base):
            …

    is equivalent to

        Base = declarative_base(name='Base')

        class MyClass(Base):
            __table_args__ = {
                'mysql_engine': 'InnoDB',
                'mysql_default charset': 'utf8',
            }
            …
    '''

    class _TableArgsMeta(declarative.DeclarativeMeta):

        def __init__(cls, name, bases, dict_):
            if (    # Do not extend base class
                    '_decl_class_registry' not in cls.__dict__ and 
                    # Missing __tablename_ or equal to None means single table
                    # inheritance — no table for it (columns go to table of
                    # base class)
                    cls.__dict__.get('__tablename__') and
                    # Abstract class — no table for it (columns go to table[s]
                    # of subclass[es]
                    not cls.__dict__.get('__abstract__', False)):
                ta = getattr(cls, '__table_args__', {})
                if isinstance(ta, dict):
                    ta = dict(table_args, **ta)
                    cls.__table_args__ = ta
                else:
                    assert isinstance(ta, tuple)
                    if ta and isinstance(ta[-1], dict):
                        tad = dict(table_args, **ta[-1])
                        ta = ta[:-1]
                    else:
                        tad = dict(table_args)
                    cls.__table_args__ = ta + (tad,)
            super(_TableArgsMeta, cls).__init__(name, bases, dict_)

    return _TableArgsMeta


table_args_meta = deprecated('Use TableArgsMeta instead.')(TableArgsMeta)



from iktomi.forms.convs import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.forms.convs is deprecated. Use ' # pragma: no cover
              'iktomi.forms.convs instead',
              category=DeprecationWarning,
              stacklevel=2)



# -*- coding: utf-8 -*-

__all__ = ['ask']

from webob import Request
from .reverse import Reverse
from .app import AppEnvironment, Application
from ..utils.storage import VersionedStorage


def ask(handler, url, method=None, data=None,
        headers=None, additional_env=None, additional_data=None,
        env_class=None):
    if isinstance(handler, Application):
        env_class = env_class or handler.env_class
        handler = handler.handler

    env_class = env_class or AppEnvironment
    root = Reverse.from_handler(handler)
    rq_kw = dict(method=method.upper()) if method else {}
    request = Request.blank(url, POST=data, headers=headers, **rq_kw)
    env = env_class.create(request, root, **(additional_env or {}))
    #TODO: may be later process cookies separatly
    data = VersionedStorage(**(additional_data or {}))
    return handler(env, data)



'''
Data classes representing references to files in model objects. Manager class
for common operations with files. Manager encapsulate knowledge on where and
how to store transient and persistent files.
'''

import os
import base64
import errno
import mimetypes
from shutil import copyfileobj
from ..utils import cached_property
import logging

logger = logging.getLogger(__name__)


class BaseFile(object):

    def __init__(self, root, name, manager=None):
        '''@root depends on environment of application and @name uniquely
        identifies the file.'''
        self.root = root
        self.name = name
        self.manager = manager

    @property
    def path(self):
        return os.path.join(self.root, self.name)

    @property
    def mimetype(self):
        '''Guessed mimetype'''
        return mimetypes.guess_type(self.path)[0]

    @cached_property
    def size(self):
        try:
            return os.path.getsize(self.path)
        # Return None for non-existing file.
        # There can be OSError or IOError (depending on Python version?), both
        # are derived from EnvironmentError having errno property.
        except EnvironmentError as exc:
            if exc.errno!=errno.ENOENT:
                raise # pragma: no cover

    @property
    def file_name(self):
        return os.path.split(self.name)[1]

    @property
    def ext(self):
        return os.path.splitext(self.name)[1]

    def __repr__(self):
        return '{}({!r})'.format(type(self).__name__, self.name)


class TransientFile(BaseFile):

    mode = 'transient'

    @property
    def url(self):
        return self.manager.get_transient_url(self)


class PersistentFile(BaseFile):

    mode = 'existing' # XXX rename existing to persistent everywhere

    @property
    def url(self):
        return self.manager.get_persistent_url(self)


def random_name(length=32):
    # altchars - do not use "-" and "_" in file names
    name = base64.b64encode(os.urandom(length), altchars=b"AA").rstrip(b'=')
    return name[:length].decode('utf-8')


class BaseFileManager(object):

    def __init__(self, persistent_root, persistent_url):
        self.persistent_root = persistent_root
        self.persistent_url = persistent_url

    def get_persistent(self, name, cls=PersistentFile):
        if not name or '..' in name or name[0] in '~/':
            raise ValueError('Unsecure file path')
        persistent = cls(self.persistent_root, name, self)
        return persistent

    def get_persistent_url(self, file, env=None):
        return self.persistent_url + file.name


class ReadonlyFileManager(BaseFileManager):
    pass


class FileManager(BaseFileManager):

    transient_length = 16
    persistent_length = 32

    def __init__(self, transient_root, persistent_root,
                 transient_url, persistent_url,
                 transient_length=None,
                 persistent_length=None):
        self.transient_root = transient_root
        self.persistent_root = persistent_root
        self.transient_url = transient_url
        self.persistent_url = persistent_url

        self.transient_length = transient_length or self.transient_length
        self.persistent_length = persistent_length or self.persistent_length

    def delete(self, file_obj):
        if os.path.isfile(file_obj.path):
            try:
                os.unlink(file_obj.path)
            except OSError as exc:
                msg = 'ERROR: {} while deleting file {}'.format(str(exc), file_obj.path)
                logger.error(msg)
                raise
        else:
            msg = 'file {} was not found while deleting'.format(file_obj.path)
            logger.warning(msg)

    def _copy_file(self, inp, path, length=None):
        # works for ajax file upload
        # XXX implement/debug for FieldStorage and file
        with open(path, 'wb') as fp:
            if length is None:
                copyfileobj(inp, fp)
            else:
                # copyfileobj does not work on request.input_stream
                # XXX check
                pos, bufsize = 0, 16*1024
                while pos < length:
                    bufsize = min(bufsize, length-pos)
                    data = inp.read(bufsize)
                    fp.write(data)
                    assert bufsize == len(data)
                    pos += bufsize

    def create_transient(self, input_stream, original_name, length=None):
        '''Create TransientFile and file on FS from given input stream and 
        original file name.'''
        ext = os.path.splitext(original_name)[1]
        transient = self.new_transient(ext)
        if not os.path.isdir(self.transient_root):
            os.makedirs(self.transient_root)

        self._copy_file(input_stream, transient.path, length=length)
        return transient

    def new_transient(self, ext=''):
        '''Creates empty TransientFile with random name and given extension.
        File on FS is not created'''
        name = random_name(self.transient_length) + ext
        return TransientFile(self.transient_root, name, self)

    def get_transient(self, name):
        '''Restores TransientFile object with given name.
        Should be used when form is submitted with file name and no file'''
        # security checks: basically no folders are allowed
        assert not ('/' in name or '\\' in name or name[0] in '.~')
        transient = TransientFile(self.transient_root, name, self)
        if not os.path.isfile(transient.path):
            raise OSError(errno.ENOENT, 'Transient file has been lost',
                          transient.path)
        return transient

    def store(self, transient_file, persistent_file):
        '''Makes PersistentFile from TransientFile'''
        #for i in range(5):
        #    persistent_file = PersistentFile(self.persistent_root,
        #                                     persistent_name, self)
        #    if not os.path.exists(persistent_file.path):
        #        break
        #else:
        #    raise Exception('Unable to find free file name')
        dirname = os.path.dirname(persistent_file.path)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)
        os.rename(transient_file.path, persistent_file.path)
        return persistent_file

    def get_transient_url(self, file, env=None):
        return self.transient_url + file.name

    def new_file_name(self, name_template, inst, ext, old_name):
        assert '{random}' in name_template, \
               'Non-random name templates are not supported yet'
        for i in range(5):
            name = name_template.format(item=inst,
                    random=random_name(self.persistent_length))
            name = name + ext
            # XXX Must differ from old value[s].
            if name != old_name or not '{random}' in name_template:
                return name
        raise Exception('Unable to find new file name') # pragma: no cover, very rare case

    def create_symlink(self, source_file, target_file):
        source_path = os.path.normpath(source_file.path)
        target_path = os.path.normpath(target_file.path)
        assert target_path.startswith(self.persistent_root), \
              'Target file must be in %s folder' % self.persistent_root
        target_dir = os.path.dirname(target_path)
        source_path_rel = os.path.relpath(source_path, target_dir)
        if not os.path.isdir(target_dir):
            os.makedirs(target_dir)
        if os.path.islink(target_path):
            os.unlink(target_path)
        os.symlink(source_path_rel, target_path)




# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)


class RouteState(object):
    def __init__(self, request):
        self._prefixes = ()
        # matched subdomain with aliases replaced by their main value
        self.primary_subdomains = () # tuple to be sure it's readonly
        self.primary_domain = ''
        # remaining subdomain part for match
        self._domain = request.host.split(':', 1)[0].encode('utf-8').decode('idna')
        self.subdomain = self._domain
        self.request = request

    def __copy__(self):
        copy = object.__new__(type(self))
        copy.__dict__.update(self.__dict__)
        return copy

    def add_prefix(self, prefix):
        self = self.__copy__()
        self._prefixes += (prefix,)
        return self

    def add_subdomain(self, subdomain, alias_matched):
        self = self.__copy__()
        if subdomain:
            self.primary_subdomains = (subdomain, ) + self.primary_subdomains
            self.primary_domain = '.'.join(self.primary_subdomains)
        if alias_matched:
            self.subdomain = self.subdomain[:-len(alias_matched)].rstrip('.')
        return self

    @property
    def path(self):
        path = self.request.path
        if self._prefixes:
            length = sum(map(len, self._prefixes))
            path = path[length:]
        return path




import re


whitespace_re = re.compile('\s+')

def pare(text, size, etc='...'):
    '''Pare text to have maximum size and add etc to the end if it's
    changed'''
    size = int(size)
    text = text.strip()
    if len(text)>size:
        # strip the last word or not
        to_be_stripped = not whitespace_re.findall(text[size-1:size+2])

        text = text[:size]
        if to_be_stripped:
            half = size//2
            last = None
            for mo in whitespace_re.finditer(text[half:]):
                last = mo

            if last is not None:
                text = text[:half+last.start()+1]

        return text.rstrip() + etc
    else:
        return text



import six
import inspect, weakref
from sqlalchemy.ext.declarative import declared_attr
from iktomi.unstable.utils.functools import return_locals


def declared_mixin(*bases):
    '''Create mix-in class with all assignments turned into methods decorated
    with declared_attr. Usage:

        @declared_mixin
        def FactoryFunction():

    or with base mix-in class[es]:

        @declared_mixin(BaseMixIn1, BaseMixIn2)
        def FactoryFunction():

    For example:

        @declared_mixin
        def WithParent():
            parent_id = Column(ForeignKey(Parent.id))
            parent = relationship(Parent)

    is equivalent to

        class WithParent(object):
            @declared_attr
            def parent_id(cls):
                return Column(ForeignKey(Parent.id))
            @declared_attr
            def parent(cls):
                return relationship(Parent)
    '''

    def wrapper(func):
        attrs = weakref.WeakKeyDictionary()
        def create_descriptor(name):
            def get_attr(cls):
                if cls not in attrs:
                    # Call func only once per class
                    attrs[cls] = return_locals(func)()
                return attrs[cls][name]
            get_attr.__name__ = name
            return declared_attr(get_attr)
        dict_ = {name: create_descriptor(name)
                 for name in six.get_function_code(func).co_varnames}
        dict_['__doc__'] = func.__doc__
        return type(func.__name__, bases, dict_)

    if len(bases)==1 and not isinstance(bases[0], type):
        # Short form (without bases) is used
        func = bases[0]
        bases = ()
        return wrapper(func)
    else:
        return wrapper



# -*- coding: utf-8 -*-
from webob.multidict import MultiDict
import six

from . import convs
from .perms import DEFAULT_PERMISSIONS
from .fields import FieldBlock


class FormValidationMetaClass(type):
    '''
    Metaclass to assert that some obsolete methods are not used.
    can be removed from iktomi after all existing code is cleaned up.
    '''

    def __new__(mcs, name, bases, dict_):
        if any([x.startswith('clean__') for x in dict_]):
            raise TypeError('Form clean__ methods are obsolete')
        return type.__new__(mcs, name, bases, dict_)


class Form(six.with_metaclass(FormValidationMetaClass, object)):

    template = 'forms/default'
    permissions = DEFAULT_PERMISSIONS
    id = ''

    def __init__(self, env=None, initial=None, name=None, permissions=None):
        initial = initial or {}
        self.env = env
        self.name = name
        self.raw_data = MultiDict()
        # NOTE: `initial` is used to set initial display values for fields.
        #       If you provide initial value for some aggregated field
        #       you need to provide values for all fields that are in that
        #       aggregated field, including `None` as empty values.
        self.initial = initial
        self.python_data = initial.copy()
        # clone all fields
        self.fields = [field(parent=self) for field in self.fields]

        if permissions is None:
            permissions = self.permissions
        self.permissions = set(permissions)

        for field in self.fields:
            # NOTE: we do not put `get_initial()` call result in `self.initial`
            #       because it may differ for each call
            self.python_data.update(field.load_initial(initial, self.raw_data))
        self.errors = {}

    @property
    def form(self):
        return self

    @property
    def prefix(self):
        '''A prefix for names of field inputs'''
        if self.name:
            return self.name+':'
        else:
            return ''

    def render(self):
        '''Proxy method to form's environment render method'''
        return self.env.template.render(self.template, form=self)

    @property
    def is_valid(self):
        '''Is true if validated form as no errors'''
        return not self.errors

    def accept(self, data):
        '''
        Try to accpet MultiDict-like object and return if it is valid.
        '''
        self.raw_data = MultiDict(data)
        self.errors = {}
        for field in self.fields:
            if field.writable:
                self.python_data.update(field.accept())
            else:
                for name in field.field_names:
                    # readonly field
                    subfield = self.get_field(name)
                    value = self.python_data[subfield.name]
                    subfield.set_raw_value(self.raw_data, subfield.from_python(value))
        return self.is_valid

    def get_field(self, name):
        '''
        Gets field by input name
        '''
        names = name.split('.', 1)
        for field in self.fields:
            if isinstance(field, FieldBlock):
                result = field.get_field(name)
                if result is not None:
                    return result
            if field.name == names[0]:
                if len(names) > 1:
                    return field.get_field(names[1])
                return field
        return None

    def get_data(self, compact=True):
        '''
        Returns data representing current state of the form. While
        Form.raw_data may contain alien fields and invalid data, this method
        returns only valid fields that belong to this form only. It's designed
        to pass somewhere current state of the form (as query string or by
        other means).
        '''
        data = MultiDict()
        for field in self.fields:
            raw_value = field.from_python(self.python_data[field.name])
            field.set_raw_value(data, raw_value)
        if compact:
            data = MultiDict([(k, v) for k, v in data.items() if v])
        return data

    def get_help(self):
        return None



# -*- coding: utf-8 -*-

from __future__ import print_function
import sys
import six
from sqlalchemy.types import SchemaType
from sqlalchemy.engine import reflection
from sqlalchemy.schema import (
    MetaData,
    Table,
    DropTable,
    ForeignKeyConstraint,
    DropConstraint,
)
from .base import Cli

__all__ = ['Sqla']


def drop_everything(engine):
    '''Droping all tables and custom types (enums) using `engine`.
    Taken from http://www.sqlalchemy.org/trac/wiki/UsageRecipes/DropEverything

    This method is more robust than `metadata.drop_all(engine)`. B.c. when
    you change a table or a type name, `drop_all` does not consider the old one.
    Thus, DB holds some unused entities.'''
    conn = engine.connect()
    # the transaction only applies if the DB supports
    # transactional DDL, i.e. Postgresql, MS SQL Server
    trans = conn.begin()
    inspector = reflection.Inspector.from_engine(engine)
    metadata = MetaData()
    tbs = []
    all_fks = []
    types = []
    for table_name in inspector.get_table_names():
        fks = []
        for fk in inspector.get_foreign_keys(table_name):
            if not fk['name']:
                continue
            fks.append(ForeignKeyConstraint((), (), name=fk['name']))
        for col in inspector.get_columns(table_name):
            if isinstance(col['type'], SchemaType):
                types.append(col['type'])
        t = Table(table_name, metadata, *fks)
        tbs.append(t)
        all_fks.extend(fks)
    try:
        for fkc in all_fks:
            conn.execute(DropConstraint(fkc))
        for table in tbs:
            conn.execute(DropTable(table))
        for custom_type in types:
            custom_type.drop(conn)
        trans.commit()
    except: # pragma: no cover
        trans.rollback()
        raise


class Sqla(Cli):
    '''
    SQLAlchemy database handling

    :param session_maker: sqlalchemy session maker function
    :param metadata: sqlalchemy metadata object or dictionary mapping names to
        metadata objects for multi-DB configuration to deal with
    :param initial: a function acceptind sqlalchemy session and filling-in
        a database with default initial data
    :param dict generators: a dictionary with generator functions. Generator
        functions should fill database with "lorem ipsum" data.
        They accept sqlalchemy session and a cnumber of objects to be created.
    '''

    def __init__(self, session_maker, metadata, initial=None, generators=None):
        self.session = session_maker()
        self.metadata = metadata
        self.initial = initial
        self.generators = generators or {}

    def _schema(self, table):
        from sqlalchemy.schema import CreateTable
        engine = self.session.get_bind(clause=table)
        return six.text_type(CreateTable(table, bind=engine))

    def command_create_tables(self, meta_name=None, verbose=False):
        '''
        Create tables according sqlalchemy data model.

        Is not a complex migration tool like alembic, just creates tables that
        does not exist::

            ./manage.py sqla:create_tables [--verbose] [meta_name]
        '''

        def _create_metadata_tables(metadata):
            for table in metadata.sorted_tables:
                if verbose:
                    print(self._schema(table))
                else:
                    print('  '+table.name)
                engine = self.session.get_bind(clause=table)
                metadata.create_all(bind=engine, tables=[table])

        if isinstance(self.metadata, MetaData):
            print('Creating tables...')
            _create_metadata_tables(self.metadata)
        else:
            for current_meta_name, metadata in self.metadata.items():
                if meta_name not in (current_meta_name, None):
                    continue
                print('Creating tables for {}...'.format(current_meta_name))
                _create_metadata_tables(metadata)

    def command_drop_tables(self, meta_name=None):
        '''
        Drops all tables without dropping a database::

            ./manage.py sqla:drop_tables [meta_name]
        '''
        answer = six.moves.input(u'All data will lost. Are you sure? [y/N] ')

        if answer.strip().lower()!='y':
            sys.exit('Interrupted')

        def _drop_metadata_tables(metadata):
            table = next(six.itervalues(metadata.tables), None)
            if table is None:
                print('Failed to find engine')
            else:
                engine = self.session.get_bind(clause=table)
                drop_everything(engine)
                print('Done')

        if isinstance(self.metadata, MetaData):
            print('Droping tables... ', end='')
            _drop_metadata_tables(self.metadata)
        else:
            for current_meta_name, metadata in self.metadata.items():
                if meta_name not in (current_meta_name, None):
                    continue
                print('Droping tables for {}... '.format(current_meta_name),
                      end='')
                _drop_metadata_tables(metadata)

    def command_init(self):
        '''
        Runs init function::

            ./manage.py sqla:init
        '''
        if self.initial:
            self.initial(self.session)

    def command_reset(self):
        '''
        Drops all tables, creates tables and runs init function::

            ./manage.py sqla:reset
        '''
        self.command_drop_tables()
        self.command_create_tables()
        self.command_init()

    def command_schema(self, name=None):
        '''
        Prints current database schema (according sqlalchemy database model)::

            ./manage.py sqla:schema [name]
        '''
        meta_name = table_name = None
        if name:
            if isinstance(self.metadata, MetaData):
                table_name = name
            elif '.' in name:
                meta_name, table_name = name.split('.', 1)
            else:
                meta_name = name

        def _print_metadata_schema(metadata):
            if table_name is None:
                for table in metadata.sorted_tables:
                    print(self._schema(table))
            else:
                try:
                    table = metadata.tables[table_name]
                except KeyError:
                    sys.exit('Table {} is not found'.format(name))
                print(self._schema(table))

        if isinstance(self.metadata, MetaData):
            _print_metadata_schema(self.metadata)
        else:
            for current_meta_name, metadata in self.metadata.items():
                if meta_name not in (current_meta_name, None):
                    continue
                _print_metadata_schema(metadata)

    def command_gen(self, *names):
        '''
        Runs generator functions.

        Run `docs` generator function::

            ./manage.py sqla:gen docs

        Run `docs` generator function with `count=10`::

            ./manage.py sqla:gen docs:10
        '''
        if not names:
            sys.exit('Please provide generator names')
        for name in names:
            name, count = name, 0
            if ':' in name:
                name, count = name.split(':', 1)
            count = int(count)
            create = self.generators[name]
            print('Generating `{0}` count={1}'.format(name, count))
            create(self.session, count)
            self.session.commit()



# -*- coding: utf-8 -*-

__all__ = ['Reverse', 'UrlBuildingError']

from .url import URL
from .url_templates import UrlBuildingError
from ..utils import cached_property



class Location(object):
    '''
    Class representing an endpoint in the reverse url map.
    '''
    # XXX For backward compatibility, some subclasses override constructor
    fragment_builder = None

    def __init__(self, *builders, **kwargs):
        self.builders = list(builders)
        self.subdomains = kwargs.get('subdomains', [])
        self.fragment_builder = kwargs.get('fragment_builder', None)

    @property
    def all_builders(self):
        builders = self.builders
        if self.fragment_builder is not None:
            # attention! not +=
            builders = builders + [self.fragment_builder]
        return builders

    @property
    def need_arguments(self):
        for b in self.all_builders:
            if b._url_params:
                return True
        return False

    def build_path(self, reverse, **kwargs):
        result = []
        for b in self.builders:
            result.append(b(**kwargs))
        return ''.join(result)

    def build_subdomians(self, reverse):
        subdomains = [getattr(x, 'primary', x)
                      for x in self.subdomains
                      if getattr(x, 'primary', x)]
        return u'.'.join(subdomains)

    def build_fragment(self, reverse, **kwargs):
        if self.fragment_builder is None:
            return None
        return self.fragment_builder(**kwargs)

    @property
    def url_arguments(self):
        result = set()
        for builder in self.all_builders:
            result |= set(builder._url_params)
        return result

    def __eq__(self, other):
        return isinstance(other, self.__class__) and \
               self.builders == other.builders and \
               self.subdomains == other.subdomains and \
               self.fragment_builder == other.fragment_builder

    def __repr__(self):
        args = '*{!r}'.format(self.builders)
        if self.subdomains:
            args += ', subdomains={!r}'.format(self.subdomains)
        if self.fragment_builder:
            args += ', fragment={!r}'.format(self.fragment_builder)
        return '{}({})'.format(self.__class__.__name__, args)



class Reverse(object):
    '''
    Object incapsulating reverse url map and methods needed to build urls
    by their names, namespaces and parameters.

    Usually an instance of `Reverse` can be found in `env.root`.
    '''
    def __init__(self, scope, location=None, path='', host='',
                 ready=False, need_arguments=False, bound_env=None, parent=None,
                 finalize_params=None, pending_args=None, fragment=None):
        # location is stuff containing builders for current reverse step
        # (builds url part for particular namespace or endpoint)
        self._location = location
        # scope is a dict having nested namespace and endpoint names as key and
        # (location, nested scope) tuple as values for the current namespace
        self._scope = scope
        self._path = path
        self._host = host
        self._fragment = fragment
        # ready means that self._location path and subdomain have been already
        # added to self._path and self._host
        self._ready = ready

        # in the case it is endpoint and
        # default rule accepts arguments, it is still callable
        self._callable = not ready or (
                    '' in scope and scope[''][0].need_arguments)

        self._need_arguments = need_arguments
        self._is_endpoint = (not self._scope) or ('' in self._scope)
        self._is_scope = bool(self._scope)
        self._bound_env = bound_env
        self._parent = parent
        self._finalize_params = finalize_params or {}
        self._pending_args = pending_args or {}

    def _attach_subdomain(self, host, location):
        subdomain = location.build_subdomians(self)
        if not host:
            return subdomain
        if subdomain:
            return subdomain + '.' + host
        return host

    def __call__(self, **kwargs):
        '''
        Get a copy of the `Reverse` but with same namespace and same url name,
        but with arguments attached.
        '''
        if not self._callable:
            raise UrlBuildingError('Endpoint do not accept arguments')
        if self._finalize_params:
            raise UrlBuildingError('Calling a reverse object multiple times is not allowed')

        if self._is_endpoint or self._need_arguments:
            finalize_params = {}
            path, host = self._path, self._host
            location = self._location
            fragment = self._fragment
            if location and not self._ready:
                kwargs.update(self._pending_args)
                host = self._attach_subdomain(host, location)
                path += location.build_path(self, **kwargs)
                # a little bit verbose to ease test coverage usage
                loc_fragment = location.build_fragment(self, **kwargs)
                if loc_fragment is not None:
                    fragment = loc_fragment
            if '' in self._scope:
                finalize_params = kwargs
            return self.__class__(self._scope, location, path=path, host=host,
                                  fragment=fragment,
                                  bound_env=self._bound_env,
                                  ready=self._is_endpoint,
                                  parent=self._parent,
                                  finalize_params=finalize_params)
        # XXX is this code reachable?
        raise UrlBuildingError('Not an endpoint {}'.format(repr(self)))

    def __getattr__(self, name):
        '''
        Get subreverse, a reverse in current namespace with the name, equal
        to the attribute name::

            env.root.index # getattr(env.root, 'index')
        '''
        if self._is_scope and name in self._scope:
            if self._need_arguments:
                return getattr(self(), name)
            location, scope = self._scope[name]
            path = self._path
            host = self._host
            fragment = self._fragment
            ready = not location.need_arguments
            if ready:
                path += location.build_path(self)
                loc_fragment = location.build_fragment(self)
                if loc_fragment is not None:
                    fragment = loc_fragment
                host = self._attach_subdomain(host, location)
            pending_args = dict(self._finalize_params)
            return self.__class__(scope, location, path, host, ready,
                                  fragment=fragment,
                                  bound_env=self._bound_env,
                                  parent=self,
                                  need_arguments=location.need_arguments,
                                  pending_args=pending_args)
        raise UrlBuildingError('Namespace or endpoint "{}" does not exist'
                               ' in {!r}'.format(name, self))

    def _finalize(self):
        # deferred build of the last part of url for endpoints that
        # also have nested scopes
        # i.e. finalization of __call__ for as_url
        if self._need_arguments:
            self = self()
        path, host = self._path, self._host
        location = self._scope[''][0]
        host = self._attach_subdomain(host, location)
        path += location.build_path(self, **self._finalize_params)
        loc_fragment = location.build_fragment(self, **self._finalize_params)
        if loc_fragment is not None:
            fragment = loc_fragment
        else:
            fragment = self._fragment
        return self.__class__({}, self._location, path=path, host=host,
                              fragment=fragment,
                              bound_env=self._bound_env,
                              parent=self._parent,
                              ready=self._is_endpoint)

    @cached_property
    def url_arguments(self):
        args = set()
        if self._is_endpoint or self._need_arguments:
            if self._location:
                args |= self._location.url_arguments
            if self._is_endpoint and self._scope:
                args |= self._scope[''][0].url_arguments
        return args

    def _build_url_silent(self, _name, **kwargs):
        subreverse = self
        used_args = set()
        for part in _name.split('.'):
            if not subreverse._ready and subreverse._need_arguments:
                used_args |= subreverse.url_arguments
                subreverse = subreverse(**kwargs)
            subreverse = getattr(subreverse, part)
        if not subreverse._ready and subreverse._is_endpoint:
            used_args |= subreverse.url_arguments
            subreverse = subreverse(**kwargs)
        return used_args, subreverse

    def build_subreverse(self, _name, **kwargs):
        '''
        String-based reverse API. Returns subreverse object::

            env.root.build_subreverse('user', user_id=1).profile
        '''
        _, subreverse = self._build_url_silent(_name, **kwargs)
        return subreverse

    def build_url(self, _name, **kwargs):
        '''
        String-based reverse API. Returns URL object::

            env.root.build_url('user.profile', user_id=1)

        Checks that all necessary arguments are provided and all
        provided arguments are used.
        '''
        used_args, subreverse =  self._build_url_silent(_name, **kwargs)

        if set(kwargs).difference(used_args):
            raise UrlBuildingError(
                'Not all arguments are used during URL building: {}'\
                    .format(', '.join(set(kwargs).difference(used_args))))
        return subreverse.as_url

    @property
    def as_url(self):
        '''
        Reverse object converted to `web.URL`.

        If Reverse is bound to env:
            * try to build relative URL,
            * use current domain name, port and scheme as default
        '''
        if '' in self._scope:
            return self._finalize().as_url

        if not self._is_endpoint:
            raise UrlBuildingError('Not an endpoint {}'.format(repr(self)))

        if self._ready:
            path, host = self._path, self._host
        else:
            return self().as_url

        # XXX there is a little mess with `domain` and `host` terms
        if ':' in host:
            domain, port = host.split(':')
        else:
            domain = host
            port = None

        if self._bound_env:
            request = self._bound_env.request
            scheme_port = {'http': '80',
                           'https': '443'}.get(request.scheme, '80')

            # Domain to compare with the result of build.
            # If both values are equal, domain part can be hidden from result.
            # Take it from route_state, not from env.request, because
            # route_state contains domain values with aliased replaced by their
            # primary value
            primary_domain = self._bound_env._route_state.primary_domain
            host_split = request.host.split(':')
            request_domain = host_split[0]
            request_port = host_split[1] if len(host_split) > 1 else scheme_port
            port = port or request_port

            return URL(path, host=domain or request_domain,
                       port=port if port != scheme_port else None,
                       scheme=request.scheme, fragment=self._fragment,
                       show_host=host and (domain != primary_domain \
                                           or port != request_port))
        return URL(path, host=domain, port=port,
                   fragment=self._fragment, show_host=True)

    def __str__(self):
        '''URLencoded representation of the URL'''
        return str(self.as_url)

    @classmethod
    def from_handler(cls, handler):
        '''
        Get unbound instance of the class related to given handler::

            app = web.cases(..)
            Reverse.from_handler(app)
        '''
        return cls(handler._locations())

    def bind_to_env(self, bound_env):
        '''
        Get a copy of the reverse, bound to `env` object.
        Can be found in env.root attribute::

            # done in iktomi.web.app.Application
            env.root = Reverse.from_handler(app).bind_to_env(env)
        '''
        return self.__class__(self._scope, self._location,
                              path=self._path, host=self._host,
                              fragment=self._fragment,
                              ready=self._ready,
                              need_arguments=self._need_arguments,
                              finalize_params=self._finalize_params,
                              parent=self._parent,
                              bound_env=bound_env)

    def __repr__(self):
        return '{}(path=\'{}\', host=\'{}\')'.format(
                self.__class__.__name__, self._path, self._host)



# -*- coding: utf-8 -*-

__all__ = ['match', 'method', 'static_files', 'prefix', 
           'subdomain', 'namespace', 'by_method']

import six
import logging
import os
from os import path
from six.moves.urllib.parse import unquote
from webob.exc import HTTPMethodNotAllowed, HTTPNotFound
from webob.static import FileApp
from .core import WebHandler, cases
from . import Response
from .url_templates import UrlTemplate
from .reverse import Location
from iktomi.utils.deprecation import deprecated


logger = logging.getLogger(__name__)


def update_data(data, new_data):
    for k,v in new_data.items():
        setattr(data, k, v)


class match(WebHandler):
    '''
    Checks if current request path completely matches given `url` pattern.
    If there are any `web.prefix` handlers in the chain before current,
    all matched prefixes are taken into account.

    Adds an endpoint to reverse url map with given `name` under the
    current namespace.
    If the name is empty, url name is equal to current namespace name::

        web.match('/', 'index') | index,
        web.match('/by-date/<date:dt>', 'by_date',
                  convs={'date': MyDateUrlConv}) | by_date,
    '''


    def __init__(self, url='', name='', fragment=None, convs=None):
        self.url = url
        self.url_name = name
        self.fragment = fragment
        self.builder = UrlTemplate(url, converters=convs)

        self.fragment = fragment
        if self.fragment is not None:
            self.fragment_builder = UrlTemplate(self.fragment, converters=convs)
        else:
            self.fragment_builder = None

    def match(self, env, data):
        matched, kwargs = self.builder.match(env._route_state.path, env=env)
        if matched is not None:
            env.current_url_name = self.url_name
            update_data(data, kwargs)
            return self.next_handler(env, data)
        return None
    __call__ = match # for beautiful tracebacks

    def _locations(self):
        location = Location(self.builder,
                            fragment_builder=self.fragment_builder)
        return {self.url_name: (location, {})}

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__,
                                       self.url, self.url_name)


class prefix(WebHandler):
    '''
    Checks if current request path matches given `url` pattern.
    It can remain an unmatched url part after mathced one.
    If there are any `web.prefix` handlers in the chain before current,
    all matched prefixes are taken into account.

    If name argument is provided, it is equal to::

        web.prefix(...) | web.namespace(name)

    Sample usage::

        web.prefix('/news', name='news') | web.cases(
            web.match() | index,
            web.match('/page/<int:page>', 'page') | page
            ),
    '''

    def __init__(self, _prefix, convs=None, name=None):
        self.url = _prefix
        self.builder = UrlTemplate(_prefix, match_whole_str=False, 
                                   converters=convs)
        if name is not None:
            # A shortcut for prefix(..) | namespace(name)
            self._next_handler = namespace(name)

    def prefix(self, env, data):
        matched, kwargs = self.builder.match(env._route_state.path, env=env)
        if matched is not None:
            update_data(data, kwargs)
            env._route_state = env._route_state.add_prefix(matched)
            result = self.next_handler(env, data)
            if result is not None:
                return result
        return None
    __call__ = prefix

    def _locations(self):
        locations = WebHandler._locations(self)
        for location, scope in locations.values():
            location.builders.insert(0, self.builder)
        return locations

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.url)


class namespace(WebHandler):
    '''
    Wraps all next handlers into url namespace with given name.
    '''

    def __init__(self, ns):
        if not ns:
            raise TypeError('Namespace must not be empty')
        # namespace is str
        self.namespace = ns

    def namespace(self, env, data):
        if hasattr(env, 'namespace'):
            env.namespace += '.' + self.namespace
        else:
            env.namespace = self.namespace
        return self.next_handler(env, data)
    __call__ = namespace

    def _locations(self):
        locations = WebHandler._locations(self)
        all_locations = [x[0] for x in locations.values()]

        # extract all common builders and subdomains from nested locations
        # and put them into namespace's location.
        # This allows to write prefixes and subdomains after namespaces:
        # prefix() | namespace() | prefix() | match()
        builders = []
        while all_locations and all_locations[0].builders:
            builder = all_locations[0].builders[0]
            if len(all_locations) > 1 and \
                not any(not (x.builders and x.builders[0] is builder)
                        for x in all_locations[1:]):
                builders.append(builder)
                for loc in all_locations:
                    loc.builders.pop(0)
            else:
                break

        # XXX do we need to do the same with subdomains?
        #     It makes everything more clear, but does not make an effect,
        #     because we have no pattern matching here
        subdomains = []
        while all_locations and all_locations[0].subdomains:
            subdomain = all_locations[0].subdomains[-1]
            if any(x.subdomains and x.subdomains[-1] is subdomain
                   for x in all_locations[1:]):
                subdomains.insert(0, subdomain)
                for loc in all_locations:
                    loc.subdomains.pop()
            else:
                break

        namespaces = self.namespace.split('.')
        for ns in namespaces[::-1]:
            locations = {ns: (Location(), locations)}

        for loc in locations.values():
            loc[0].builders = builders
            loc[0].subdomains = subdomains

        return locations


class method(WebHandler):

    '''Checks whether request method is in a list of allowed methods.

    If `strict=True`, raises `webob.exc.HTTPMethodNotAllowed` on wrong method,
    otherwise returns `None` ("continue route" signal)::

        web.method('GET', 'POST', strict=True) | handler
    '''

    def __init__(self, *names, **kw):
        self._names = set([name.upper() for name in names])
        if 'GET' in self._names:
            self._names.add('HEAD')
        self.strict = kw.pop('strict', False)
        assert not kw

    def method(self, env, data):
        if env.request.method in self._names:
            return self.next_handler(env, data)
        if self.strict:
            raise HTTPMethodNotAllowed()
        return None
    __call__ = method

    def __repr__(self):
        return 'method({})'.format(', '.join(repr(n) for n in self._names))


class by_method(cases):
    '''
    Cases chooser handler::

        by_method({'GET': get_item_handler,
                   'POST': save_item_handler})
    '''

    def __init__(self, handlers_dict, default_handler=None):
        handlers = []
        for methods, handler in handlers_dict.items():
            if isinstance(methods, six.string_types):
                methods = (methods,)
            handlers.append(method(*methods, strict=False) | handler)
        if default_handler is not None:
            handlers.append(default_handler)
        else:
            handlers.append(HTTPMethodNotAllowed())
        cases.__init__(self, *handlers)


class subdomain(WebHandler):
    '''
    Checks if current request domain matches given one.

    `subdomains`: a list of following values in order they are
    attempted to match:

        * string — domain part (without dots after and before)
        * empty string — no pattern is matched and there must not
          remain any unmatched domain part, subdomain matching is finished
        * `None` — no pattern is matched and but there can remain
          an unmatched domain part, subdomain matching can
          continue in next handlers

    `name`: shortcut for namespace handler chained next to subdomain:
        ::

            web.subdomain(...) | web.namespace(name)

    `primary`: one of subdomains, used to reverse an url with this 
    subdomain from other domains. 
    All other subdomains are aliases to primary subdomain.
    By default, the first subdomain.
    '''

    def __init__(self, *subdomains, **kwargs):
        self.subdomains = subdomains

        # this attribute is used for ducktyping in Location, be careful
        self.primary = kwargs.pop('primary', self.subdomains[0])
        assert self.primary in self.subdomains

        name = kwargs.pop('name', None)
        if name is not None:
            # A shortcut for subdomain(..) | namespace(name)
            self._next_handler = namespace(name)

        if kwargs: # pragma: no cover
            raise TypeError("subdomain.__init__ got an unexpected keyword "
                            "arguments {}".format(",".join(kwargs)))

    def subdomain(self, env, data):
        subdomain = env._route_state.subdomain
        #XXX: here we can get 'idna' encoded sequence, that is the bug
        for subd in self.subdomains:
            if subd:
                slen = len(subd)
                delimiter = subdomain[-slen-1:-slen]
                matches = subdomain.endswith(subd) and delimiter in ('', '.')
            elif subd is None:
                # continue matching 
                matches = True
            elif subd == '':
                # no subdomains are allowed
                matches = not subdomain

            if matches:
                env._route_state = \
                        env._route_state.add_subdomain(self.primary, subd)
                return self.next_handler(env, data)
        return None
    __call__ = subdomain

    def _locations(self):
        locations = WebHandler._locations(self)
        for location, scope in locations.values():
            location.subdomains.append(self)
        return locations

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.subdomains)


class static_files(WebHandler):
    '''
    Static file handler for dev server (not recommended in production)::

       static_files('/path/to/static', url='/static/')
    '''

    def __init__(self, location, url='/static/'):
        self.location = location
        self.url = url

    def url_for_static(self, part):
        while part.startswith('/'):
            part = part[1:]
        return path.join(self.url, part)

    @deprecated("Use static.url_for_static instead")
    def construct_reverse(self): # pragma: no cover
        return self.url_for_static

    def translate_path(self, pth):
        """Translate a /-separated PATH to the local filename syntax."""
        # initially copied from SimpleHTTPServer
        words = pth.split('/')
        words = filter(None, words)
        pth = self.location
        for word in words:
            # Do not allow path separators other than /,
            # drive names and . ..
            drive, word = os.path.splitdrive(word)
            head, word = os.path.split(word)
            if drive or head or word in (os.curdir, os.pardir):
                return None
            pth = os.path.join(pth, word)

        assert pth.startswith(self.location + '/')
        assert pth == path.normpath(pth)
        return pth

    def static_files(self, env, data):
        path_info = unquote(env.request.path)
        if path_info.startswith(self.url):
            if path_info.endswith('/'):
                raise HTTPNotFound
            file_path = self.translate_path(path_info[len(self.url):])
            if file_path and path.exists(file_path) and path.isfile(file_path):
                return FileApp(file_path)
            else:
                logger.info('Client requested non existent static data "%s"',
                            file_path)
                return Response(status=404)
        return None
    __call__ = static_files

    def __repr__(self):
        return '{}({!r}, {!r})'.format(self.__class__.__name__, 
                                       self.location, self.url)




# i18n markers
def N_(msg):
    '''
    Single translatable string marker.
    Does nothing, just a marker for \\*.pot file compilers.

    Usage::

        n = N_('translate me')
        translated = env.gettext(n)
    '''
    return msg


class M_(object):
    '''
    Marker for translatable string with plural form.
    Does not make a translation, just incapsulates a data about
    the translatable string.

    :param single: a single form
    :param plural: a plural form. Count can be included in %\-format syntax
    :param count_field: a key used to format

    Usage::

        message = M_(u'max length is %(max)d symbol',
                     u'max length is %(max)d symbols',
                     count_field="max")
        m = message % {'max': 10}
        trans = env.ngettext(m.single,
                             m.plural,
                             m.count
                             ) % m.format_args
    '''
    def __init__(self, single, plural, count_field='count', format_args=None):
        self.single = single
        self.plural = plural
        self.count_field = count_field
        self.format_args = format_args

    def __mod__(self, format_args):
        '''
        Returns a copy of the object with bound formatting args (as dict).
        A key equal to `count_field` must be in `format_args`.
        '''
        return self.__class__(self.single, self.plural, count_field=self.count_field,
                              format_args=format_args)

    @property
    def count(self):
        '''
        A count based on `count_field` and `format_args`.
        '''
        args = self.format_args
        if args is None or \
                (isinstance(args, dict) and self.count_field not in args):
            raise TypeError("count is required")
        return args[self.count_field] if isinstance(args, dict) else args

    def __unicode__(self):
        args = self.format_args
        if self.count == 1:
            return self.single % args
        return self.plural % args









import six
import os, errno, logging, inspect
from sqlalchemy.orm.session import object_session
from sqlalchemy.orm.interfaces import MapperProperty
from sqlalchemy.orm.attributes import get_history
from sqlalchemy.orm.util import class_mapper
from sqlalchemy import event
from sqlalchemy.util import set_creation_order
from weakref import WeakKeyDictionary
from iktomi.db.files import TransientFile, PersistentFile

logger = logging.getLogger(__name__)


class FileEventHandlers(object):

    file_name_encoding = 'utf-8'

    def __init__(self, prop):
        self.prop = prop

    def _get_history(self, target):
        return get_history(target, self.prop.attribute_name)

    @staticmethod
    def _remove_file(path):
        try:
            os.remove(path)
        except OSError as exc:
            if exc.errno==errno.ENOENT:
                logger.warning("Can't remove file %r: doesn't exist", path)
                #raise # XXX
            else:
                raise # pragma: no cover

    def _store_transient(self, target):
        transient = getattr(target, self.prop.key)
        if transient is None:
            for file_attr, target_attr in self.prop.cache_properties.items():
                setattr(target, target_attr, None)
            return
        if isinstance(transient, PersistentFile):
            return
        assert isinstance(transient, TransientFile), repr(transient)
        persistent = self._2persistent(target, transient)
        file_attr = getattr(type(target), self.prop.key)
        file_attr._states[target] = persistent

        for file_attr, target_attr in self.prop.cache_properties.items():
            setattr(target, target_attr, getattr(persistent, file_attr))

    def _2persistent(self, target, transient):
        session = object_session(target)
        persistent_name = getattr(target, self.prop.attribute_name)
        persistent_name = self._decode_file_name(persistent_name)
        attr = getattr(type(target), self.prop.key)
        file_manager = session.find_file_manager(attr)
        persistent = file_manager.get_persistent(persistent_name,
                                                 self.prop.persistent_cls)

        file_attr = getattr(target.__class__, self.prop.key)
        file_manager = session.find_file_manager(file_attr)

        return file_manager.store(transient, persistent)

    def _decode_file_name(self, file_name):
        if isinstance(file_name, bytes):
            return file_name.decode(self.file_name_encoding)
        else:
            return file_name

    def before_insert(self, mapper, connection, target):
        self._store_transient(target)

    def before_update(self, mapper, connection, target):
        changes = self._get_history(target)
        if not (changes.deleted or changes.added):
            return
        if changes.deleted:
            old_name = self._get_file_name_to_delete(target, changes)
            if old_name:
                old_name = self._decode_file_name(old_name)
                session = object_session(target)

                file_attr = getattr(target.__class__, self.prop.key)
                file_manager = session.find_file_manager(file_attr)

                old = file_manager.get_persistent(old_name,
                                                  self.prop.persistent_cls)
                self._remove_file(old.path)
        self._store_transient(target)

    def _get_file_name_to_delete(self, target, changes):
        if changes and changes.deleted:
            file_name = changes.deleted[0]
            if file_name is not None:
                return self._decode_file_name(file_name)

    def after_delete(self, mapper, connection, target):
        changes = self._get_history(target)
        old_name = self._get_file_name_to_delete(target, changes)
        old_name = old_name or getattr(target, self.prop.attribute_name)
        if old_name is not None:
            old_name = self._decode_file_name(old_name)
            session = object_session(target)

            file_attr = getattr(target.__class__, self.prop.key)
            file_manager = session.find_file_manager(file_attr)

            old = file_manager.get_persistent(old_name,
                                              self.prop.persistent_cls)
            self._remove_file(old.path)


class FileAttribute(object):

    def __init__(self, prop, class_=None):
        self.prop = prop
        self.column = prop.column
        self.attribute_name = prop.attribute_name
        self.name_template = prop.name_template
        self.class_ = class_
        self.cache_properties = prop.cache_properties
        self.persistent_cls = prop.persistent_cls
        # State for each instance
        self._states = WeakKeyDictionary()

    def __get__(self, inst, cls=None):
        if inst is None:
            return self
        if inst not in self._states:
            # XXX column name may differ from attribute name
            # empty string should be considered as None
            value = getattr(inst, self.attribute_name) or None
            if value is not None:
                session = object_session(inst)
                if session is None:
                    raise RuntimeError('Object is detached')
                if not hasattr(session, 'file_manager'):
                    raise RuntimeError(
                            "Session doesn't support file management")
                file_manager = session.find_file_manager(self)
                value = file_manager.get_persistent(value.decode('utf-8'),
                                                    self.persistent_cls)

                for file_attr, target_attr in self.cache_properties.items():
                    setattr(value, file_attr, getattr(inst, target_attr))

            self._states[inst] = value
        return self._states[inst]

    def __set__(self, inst, value):
        if inst in self._states and self._states[inst]==value:
            return

        # sqlalchemy bug workaround
        # To get correct history we should assert that old value has been 
        # loaded from database. getattr loads lazy attribute.
        # See http://www.sqlalchemy.org/trac/ticket/2787
        old_name = getattr(inst, self.attribute_name)

        self._states[inst] = value
        if value is None:
            setattr(inst, self.attribute_name, None)
        elif isinstance(value, TransientFile):
            ext = os.path.splitext(value.name)[1]
            # XXX getting manager from a file object 
            #     looks like a hack
            name = value.manager.new_file_name(
                    self.name_template, inst, ext, old_name)
            setattr(inst, self.attribute_name, name.encode('utf-8'))
        elif isinstance(value, PersistentFile):
            setattr(inst, self.attribute_name, value.name.encode('utf-8'))

            for file_attr, target_attr in self.cache_properties.items():
                setattr(inst, target_attr, getattr(value, file_attr))
        else:
            raise ValueError('File property value must be TransientFile, '\
                             'PersistentFile or None')


class FileProperty(MapperProperty):

    attribute_cls = FileAttribute
    event_cls = FileEventHandlers

    def __init__(self, column, name_template, attribute_name=None, **options):
        super(FileProperty, self).__init__()
        self.column = column
        self._attribute_name = attribute_name
        self.name_template = name_template
        self._set_options(options)
        set_creation_order(self)

    @property
    def attribute_name(self):
        return self._attribute_name or self.column.key

    def _set_options(self, options):
        self.persistent_cls = options.pop('persistent_cls', PersistentFile)
        self.cache_properties = dict(options.pop('cache_properties', {}))
        assert not options, "Got unexpeted parameters: %s" % (
                options.keys())

    def instrument_class(self, mapper):
        handlers = self.event_cls(self)
        event.listen(mapper, 'before_insert', handlers.before_insert,
                     propagate=True)
        event.listen(mapper, 'before_update', handlers.before_update,
                     propagate=True)
        event.listen(mapper, 'after_delete', handlers.after_delete,
                     propagate=True)
        setattr(mapper.class_, self.key,
                self.attribute_cls(self, mapper.class_))

    # XXX Implement merge?



def filesessionmaker(sessionmaker, file_manager, file_managers=None):
    u'''Wrapper of session maker adding link to a FileManager instance
    to session.::

        file_manager = FileManager(cfg.TRANSIENT_ROOT,
                                   cfg.PERSISTENT_ROOT)
        filesessionmaker(sessionmaker(...), file_manager)
    '''

    registry = WeakKeyDictionary()

    if file_managers:
        for k, v in six.iteritems(file_managers):
            if isinstance(k, FileAttribute):
                raise NotImplementedError()
            registry[k] = v

    def find_file_manager(self, target):
        if isinstance(target, FileAttribute):
            assert hasattr(target, 'class_')
            target = target.class_
        else:
            if not inspect.isclass(target):
                target = type(target)

        assert hasattr(target, 'metadata')
        assert class_mapper(target) is not None
        if target in registry:
            return registry[target]
        if target.metadata in registry:
            return registry[target.metadata]
        return file_manager

    def session_maker(*args, **kwargs):
        session = sessionmaker(*args, **kwargs)
        # XXX in case we want to use session manager somehow bound 
        #     to request environment. For example, to generate user-specific
        #     URLs.
        #session.file_manager = \
        #        kwargs.get('file_manager', file_manager)
        session.file_manager = file_manager
        session.find_file_manager = six.create_bound_method(
                                            find_file_manager,
                                            session)

        return session
    return session_maker



# -*- coding: utf-8 -*-
import sys
from distutils.core import setup


install_requires = ['six']
extras_requires = {
    'web': ['webob'],
    'fcgi': ['flup6'],
    'sqla': ['sqlalchemy', 'packaging'],
    'memcached': ['python-memcached'],
    'cleanhtml': ['lxml'],
    'renderhtml': ['jinja2'],
    'images': ['pillow'],
}

tests_requires = [
    'pymysql',
    'pytest',
    'pytest-cov',
    'webtest',
]
if sys.version_info[0] < 3:
    tests_requires.append('mock')

dependency_links = [
    'https://github.com/ods/testalchemy/tarball/master#egg=testalchemy-0.4',
    'https://github.com/lunant/mockcache/tarball/master#egg=mockcache-1.0.3_alpha',
]

extras_requires['tests'] = tests_requires

setup(
    name='iktomi',
    version='0.14',
    packages=['iktomi',
              'iktomi.utils',
              'iktomi.forms',
              'iktomi.web',
              'iktomi.templates',
                    'iktomi.templates.jinja2',
              'iktomi.db',
                    'iktomi.db.sqla',
              'iktomi.cli',
              'iktomi.unstable',
                    'iktomi.unstable.forms',
                    'iktomi.unstable.utils',
                    'iktomi.unstable.db',
                        'iktomi.unstable.db.sqla'],
    package_dir={
        'iktomi.templates.jinja2': 'iktomi/templates/jinja2',
    },
    package_data={
        'iktomi.templates.jinja2': ['templates/*/*.html'],
    },
    install_requires=install_requires,
    extras_require=extras_requires,
    tests_require=tests_requires,
    dependency_links=dependency_links,
    author='Denis Otkidach',
    author_email='denis.otkidach@gmail.com',
    maintainer='Harut Dagesyan',
    maintainer_email='yes@harutune.name',
    description='A web tool: routing, forms, other useful things.',
    # long_description=open('README').read(),
    url='http://github.com/SmartTeleMax/iktomi/',
    license='MIT',
    keywords='web forms',
)



# -*- coding: utf-8 -*-

__all__ = ['URL']

import six
if six.PY2:
    from urlparse import urlparse, parse_qs, unquote
else:# pragma: no cover; we check coverage only in python2 part
    from urllib.parse import urlparse, parse_qs, unquote
from webob.multidict import MultiDict
from .url_templates import urlquote
from iktomi.utils.url import uri_to_iri_parts


def construct_url(path, query, host, port, scheme, fragment=None):
    query = ('?' + '&'.join('{}={}'.format(urlquote(k), urlquote(v))
                            for k, v in six.iteritems(query))
             if query else '')

    hash_part = ('#' + fragment) if fragment is not None else ''

    if host:
        host = host.encode('idna').decode('utf-8')
        port = ':' + port if port else ''
        return ''.join((scheme, '://', host, port, path, query, hash_part))
    else:
        return path + query + hash_part

if six.PY2:
    def _parse_qs(query):
        return sum([[(k.decode('utf-8', errors="replace"),
                      v.decode('utf-8', errors="replace"))
                     for v in values]
                    for k, values in parse_qs(query).items()], [])

    def _unquote(path):
        # in PY2 unquote returns encoded value of the type it has accepted
        return unquote(path.encode('utf-8')).decode('utf-8')
else:# pragma: no cover
    def _parse_qs(query):
        return sum([[(k, v) for v in values]
                     for k, values in parse_qs(query).items()], [])

    # in PY3 is accepts and returns decoded str
    _unquote = unquote


def _decode_path(path):
    if path is None:
        return None
    if isinstance(path, six.binary_type):
        path = path.decode('utf-8', errors="replace") # XXX
    path = urlquote(path)
    return path


def safe_idna(s):
    try:
        return s.encode('idna').decode('idna')
    except UnicodeError:
        return s


class URL(str):

    # attributes stored in the object, as for now:
    #     path - urlencoded string of text_type (not bytes)
    #     host - unicode idna-decoded
    #     query - dict of unicode keys and unicode or implementing
    #             string convertion values
    #     fragment - None or urlencoded string of text_type

    def __new__(cls, path=None, query=None, host=None, port=None, scheme=None,
                fragment=None, show_host=True, uri_path=None, uri_fragment=None):
        '''
        path - url path part, unicode string (not encoded at all) of text_type
            or urlencoded value of binary_type.
        scheme - an url scheme, http by default
        host - host without port, idna-encoded or not encoded unicode string
        port - port number of string type
        query - dict of text_type keys and values of text_type or supporting
            convertion to text_type
        fragment - fragment (hash) part of the url,
            None (for no hash part)
            or empty string (for #)
            or unicode string (not encoded at all) of text_type
            or urlencoded value of binary_type.
        show_host - a boolean value indicating whether to show host part
            (scheme, host, port)
        uri_path, uri_fragment - same as path and fragment parameters
            but accepts only urlencoded strings of text_type without guessing
            whether it is already encoded or not.
            Useful to avoid double quotting.
            Overrides path and fragment parameters.
        '''
        path = uri_path or _decode_path(path)

        # Note: it is bad idea to use unicode in fragment part in browsers.
        #       We encode it according to RFC and Firefox does it as well,
        #       but Chrome allows unicode and does not encode/decode it at all
        #       and is uncompatible with RFC.
        fragment = uri_fragment or _decode_path(fragment)

        query = MultiDict(query) if query else MultiDict()
        host = host or ''
        port = port or ''
        scheme = scheme or 'http'
        _self = construct_url(path, query, host if show_host else '',
                              port, scheme, fragment)
        self = str.__new__(cls, _self)
        self.path = path
        self.query = query

        # force decode idna from both encoded and decoded input
        self.host = '.'.join(safe_idna(x) for x in host.split('.'))
        self.port = port
        self.scheme = scheme
        self.fragment = fragment
        self.show_host = show_host
        return self

    @classmethod
    def from_url(cls, url, show_host=True):
        '''Parse string and get URL instance'''
        # url must be idna-encoded and url-quotted

        if six.PY2:
            if isinstance(url, six.text_type):
                url = url.encode('utf-8')
            parsed = urlparse(url)
            netloc = parsed.netloc.decode('utf-8') # XXX HACK
        else:# pragma: no cover
            if isinstance(url, six.binary_type):
                url = url.decode('utf-8', errors='replace') # XXX
            parsed = urlparse(url)
            netloc = parsed.netloc

        query = _parse_qs(parsed.query)
        host = netloc.split(':', 1)[0] if ':' in netloc else netloc

        port = netloc.split(':')[1] if ':' in netloc else ''
        path = unquote(parsed.path)
        fragment = unquote(parsed.fragment)
        if not fragment and not url.endswith('#'):
            fragment = None
        return cls(path,
                   query, host,
                   port, parsed.scheme, fragment, show_host)

    def __reduce__(self):
        return (self.__class__.from_url, (str(self),))

    def _copy(self, **kwargs):
        kw = dict(query=self.query, host=self.host,
                  port=self.port, scheme=self.scheme,
                  show_host=self.show_host)
        kw.update(kwargs)
        if 'path' not in kw:
            kw['uri_path'] = self.path
        if 'fragment' not in kw:
            kw['uri_fragment'] = self.fragment
        return self.__class__(**kw)

    def qs_set(self, *args, **kwargs):
        '''Set values in QuerySet MultiDict'''
        if args and kwargs:
            raise TypeError('Use positional args or keyword args not both')
        query = self.query.copy()
        if args:
            mdict = MultiDict(args[0])
            for k in mdict.keys():
                if k in query:
                    del query[k]
            for k, v in mdict.items():
                query.add(k, v)
        else:
            for k, v in kwargs.items():
                query[k] = v
        return self._copy(query=query)

    def qs_add(self, *args, **kwargs):
        '''Add value to QuerySet MultiDict'''
        query = self.query.copy()
        if args:
            mdict = MultiDict(args[0])
            for k, v in mdict.items():
                query.add(k, v)
        for k, v in kwargs.items():
            query.add(k, v)
        return self._copy(query=query)

    def with_host(self):
        '''Force show_host parameter'''
        return self._copy(show_host=True)

    def qs_delete(self, *keys):
        '''Delete value from QuerySet MultiDict'''
        query = self.query.copy()
        for key in set(keys):
            try:
                del query[key]
            except KeyError:
                pass
        return self._copy(query=query)

    def qs_get(self, key, default=None):
        '''Get a value from QuerySet MultiDict'''
        return self.query.get(key, default=default)

    def get_readable(self):
        '''
        Gets human-readable representation of the url (as unicode string,
        IRI according RFC3987)
        '''
        query = (u'?' + u'&'.join(u'{}={}'.format(urlquote(k), urlquote(v))
                                  for k, v in six.iteritems(self.query))
                 if self.query else '')
        hash_part = (u'#' + self.fragment) if self.fragment is not None else u''

        path, query, hash_part = uri_to_iri_parts(self.path, query, hash_part)

        if self.host:
            port = u':' + self.port if self.port else u''
            return u''.join((self.scheme, '://', self.host, port, path, query, hash_part))
        else:
            return path + query + hash_part

    def __repr__(self):
        return '<URL {!r}>'.format(str(self))



# -*- coding: utf-8 -*-

import os
import sys
import time
import logging
import threading
from itertools import chain

from .base import Cli

__all__ = ['App']


logger = logging.getLogger(__name__)


try:
    MAXFD = os.sysconf("SC_OPEN_MAX")
except: # pragma: no cover
    MAXFD = 256


def flush_fds():
    for fd in range(3, MAXFD + 1):
        try:
            os.fsync(fd)
        except OSError:
            pass


class App(Cli):
    '''
    Development application

    :param app: iktomi app
    :param shell_namespace: dict with initial namespace for shell command
    :param extra_files: extra files to watch and reload if they are changed
    :param bootstrap: bootstrap function before called dev server is being runned
    '''
    format = '%(levelname)s [%(name)s] %(message)s'

    def __init__(self, app, shell_namespace=None, extra_files=None, bootstrap=None):
        self.app = app
        self.shell_namespace = shell_namespace or {}
        self.extra_files = extra_files
        self.bootstrap = bootstrap

    def command_serve(self, host='', port='8000', level='debug'):
        '''
        Run development server with automated reload on code change::

            ./manage.py app:serve [host] [port] [level]
        '''
        logging.basicConfig(level=getattr(logging, level.upper()), format=self.format)
        if self.bootstrap:
            logger.info('Bootstraping...')
            self.bootstrap()
        try:
            server_thread = DevServerThread(host, port, self.app)
            server_thread.start()

            wait_for_code_change(extra_files=self.extra_files)
            server_thread.running = False
            server_thread.join()
            logger.info('Reloading...')
            flush_fds()
            pid = os.fork()
            # We need to fork before `execvp` to perform code reload
            # correctly, because we need to complete python destructors and
            # `atexit`.
            # This will save us from problems of incorrect exit, such as:
            # - unsaved data in data storage, which does not write data
            # on hard drive immediatly
            # - code, that can't be measured with coverage tool, because it uses
            # `atexit` handler to save coverage data
            # NOTE: we using untipical fork-exec scheme with replacing
            # the parent process(not the child) to preserve PID of proccess
            # we use `pragma: no cover` here, because parent process cannot be
            # measured with coverage since it is ends with `execvp`
            if pid: # pragma: no cover
                os.closerange(3, MAXFD)
                os.waitpid(pid, 0)
                # reloading the code in parent process
                os.execvp(sys.executable, [sys.executable] + sys.argv)
            else:
                # we closing our recources, including file descriptors
                # and performing `atexit`.
                sys.exit()
        except KeyboardInterrupt:
            logger.info('Stoping dev-server...')
            server_thread.running = False
            server_thread.join()
            sys.exit()

    def command_shell(self):
        '''
        Shell command::

            ./manage.py app:shell

        Executed with `self.shell_namespace` as local variables namespace.
        '''
        from code import interact
        interact('Namespace {!r}'.format(self.shell_namespace),
                 local=self.shell_namespace)


class DevServerThread(threading.Thread):

    def __init__(self, host, port, app):
        from wsgiref.simple_server import make_server, WSGIServer, \
                WSGIRequestHandler
        self.host = host
        self.port = port

        class DevServer(WSGIServer):
            timeout = 0.2


        class RequestHandler(WSGIRequestHandler):

            def address_string(slf):
                # getfqdn sometimes is very slow
                return '{}:{}'.format(host, port)

            def log_message(self, format, *args):
                logger.info("%s - - [%s] %s",
                            self.client_address[0],
                            self.log_date_time_string(),
                            format % args)

        try:
            self.port = int(port)
        except ValueError:
            raise ValueError(
                'Please provide valid port value insted of "{}"'.format(port))
        self.running = True
        self.server = make_server(self.host, self.port, app, server_class=DevServer,
                                  handler_class=RequestHandler)
        super(DevServerThread, self).__init__()

    def run(self):
        logger.info('Devserver is running on port %s\n', self.port)
        while self.running:
            self.server.handle_request()


# All reloader utils are taken from werkzeug
def iter_module_files():
    for module in sys.modules.values():
        filename = getattr(module, '__file__', None)
        if filename:
            while not os.path.isfile(filename): # pragma: no cover
                # NOTE: this code is needed for the cases of importing
                # from archive or custom importers
                # for example, if we importing from archive foo.zip
                # module named zipped, then this zipped.__file__ will equal
                # to foo.zip/zipped.py, and os.path.dirname will give us
                # file, not directory.
                # It is marked as pragma: no cover, because this code was taken
                # from werkzeug and we believe that it is tested
                filename = os.path.dirname(filename)
                if not filename:
                    break
            else:
                if filename.endswith(('.pyc', '.pyo')):
                    filename = filename[:-1]
                yield filename


def wait_for_code_change(extra_files=None, interval=1):
    mtimes = {}
    while 1:
        for filename in chain(iter_module_files(), extra_files or ()):
            try:
                mtime = os.stat(filename).st_mtime
            except OSError: # pragma: no cover
                # this is cannot be guaranteed covered by coverage because of interpreter optimization
                # see https://bitbucket.org/ned/coveragepy/issues/198/continue-marked-as-not-covered#comment-4052311

                continue

            old_time = mtimes.get(filename)
            if old_time is None:
                mtimes[filename] = mtime
            elif mtime > old_time:
                logger.info('Changes in file "%s"', filename)
                return
        time.sleep(interval)




# -*- coding: utf-8 -*-


class Storage(object):
    def set(self, key, value, time=0):# pragma: no cover
        raise NotImplementedError()
    def get(self, key, default=None):# pragma: no cover
        raise NotImplementedError()
    def delete(self, key):# pragma: no cover
        raise NotImplementedError()


class LocalMemStorage(Storage):
    def __init__(self):
        self.storage = {}

    def set(self, key, value, time=0):
        self.storage[key] = value
        return True

    def get(self, key, default=None):
        return self.storage.get(key, default)

    def delete(self, key):
        if key in self.storage:
            del self.storage[key]
        return True


class MemcachedStorage(Storage):
    def __init__(self, conf):
        import memcache
        conf = conf if isinstance(conf, (list, tuple)) else [conf]
        self.storage = memcache.Client(conf)

    def set(self, key, value, time=0):
        return self.storage.set(key, value, time)

    def get(self, key, default=None):
        value = self.storage.get(key)
        if value is None:
            return default
        return value

    def delete(self, key):
        return self.storage.delete(key)
